#ifndef __HTTP_H__
#define __HTTP_H__

/* HTTP parsing */

/* http versions */
#define MHTTP_HTTP_VER_1_0	0
#define MHTTP_HTTP_VER_1_1	1

/* host/post */
#define MHTTP_GET		0
#define MHTTP_POST		1
#define MHTTP_CONNECT		2

#ifndef M_HTTP_T_DEFINED
typedef void *M_http_t;
#endif

#ifdef M_INTERNAL
#  include "monetra_module_setup.h"
#else
#  define M_MAIN_EXPORT
#endif

#ifdef LIBMONETRA_INTERNAL
#  define HTTP_NS(b) LIBMONETRA_##b
#else
#  define HTTP_NS(b) M_##b
#endif


/* init/destroy */
M_MAIN_EXPORT void     HTTP_NS(http_free)(M_http_t http);
M_MAIN_EXPORT M_http_t HTTP_NS(http_new)(void);

/* generate new request */
M_MAIN_EXPORT int   HTTP_NS(http_get_gettype)(M_http_t http);
M_MAIN_EXPORT char *HTTP_NS(http_gen_request)(M_http_t http);
M_MAIN_EXPORT char *HTTP_NS(http_gen_resp)(M_http_t http);
M_MAIN_EXPORT int   HTTP_NS(http_get_request_len)(M_http_t http);
M_MAIN_EXPORT int   HTTP_NS(http_get_resp_len)(M_http_t http);
M_MAIN_EXPORT int   HTTP_NS(http_get_version)(M_http_t http);
M_MAIN_EXPORT int   HTTP_NS(http_get_complete)(M_http_t http);
M_MAIN_EXPORT int   HTTP_NS(http_set_content)(M_http_t http, const char *content, int len);
M_MAIN_EXPORT int   HTTP_NS(http_set_header)(M_http_t http, const char *header, const char *value, int overwrite);
M_MAIN_EXPORT int   HTTP_NS(http_set_hosttype)(M_http_t http, int hosttype, const char *path, const char *host, unsigned short port);
M_MAIN_EXPORT int   HTTP_NS(http_set_resp_code)(M_http_t http, int code, const char *code_msg);
M_MAIN_EXPORT int   HTTP_NS(http_set_version)(M_http_t http, int version);

/* parse response */
M_MAIN_EXPORT M_http_t HTTP_NS(http_parse)(const char *buf, int len, int return_on_incomplete, char *debugmsg, int msglen);
M_MAIN_EXPORT int      HTTP_NS(http_process_chunksection)(char *data, int len, char **out_unchunked, int *out_len);
M_MAIN_EXPORT int      HTTP_NS(http_get_code)(M_http_t http, char **code_msg);
M_MAIN_EXPORT int      HTTP_NS(http_get_content)(M_http_t http, char **content);
M_MAIN_EXPORT int      HTTP_NS(http_get_content_len)(M_http_t http);
M_MAIN_EXPORT int      HTTP_NS(http_get_header_len)(M_http_t http);
M_MAIN_EXPORT int      HTTP_NS(http_get_header_int)(M_http_t http, const char *header, int idx);
M_MAIN_EXPORT char    *HTTP_NS(http_get_header_str)(M_http_t http, const char *header, int idx);

#endif
