/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#include "monetra.h"
#include "libmonetra_main.h"

/* Parse CSV data.
 * The passed in 'str' may be modified, and the parsed CSV data
 * may rely on 'str' so it should not be de-allocated within
 * the lifetime of the use of the csv array */
char ***M_parse_csv(char *str, int len, int *num_rows_out, int *num_cols_out)
{
	int num_rows = 0;
	int num_cols = 0;
	int on_quote = 0;
	char **rows = NULL;
	char *temp = NULL;
	int templen = 0;
	char ***ret = NULL;
	int i, j;
	int on_col = 0;

	(*num_rows_out) = 0;
	(*num_cols_out) = 0;

	if (!str || len == 0)
		return(NULL);

	/* -- Separate the data into rows/lines */

	/* Allocate rows in 1024 blocks */
	rows = malloc(1024 * sizeof(*rows));
	rows[num_rows++] = str;
	for (i=0; i<len; i++) {
		if (!on_quote && str[i] == '"') {
			on_quote = 1;
		} else if (on_quote && str[i] == '"') {
			if (str[i+1] == '"') {
				/* Skip past double quote */
				i++;
			} else {
				on_quote = 0;
			}
		} else if (!on_quote && (str[i] == '\r' || str[i] == '\n')) {
			str[i] = 0;
			if (!strlen(rows[num_rows-1])) {
				/* Move pointer forward */
				rows[num_rows-1] = str+(i+1);
			} else {
				/* Ensure there is enough room in our array */
				if (num_rows % 1024 == 0) {
					rows = realloc(rows, (num_rows + 1024) * sizeof(*rows));
				}
				rows[num_rows++] = str+(i+1);
			}
		}
	}

	/* Can't end inside of quoted data */
	if (on_quote) {
		free(rows);
		return(NULL);
	}

	/* No blank row at end */
	if (!strlen(rows[num_rows-1]))
		num_rows--;

	if (num_rows == 0) {
		free(rows);
		return(NULL);
	}

	/* -- Count the number of columns in first row (header row) -- */
	temp = rows[0];
	templen = strlen(temp);
	on_quote = 0;
	num_cols = 1; /* Start with 1 column */
	for (i=0; i<templen; i++) {
		if (!on_quote && temp[i] == '"') {
			on_quote = 1;
		} else if (on_quote && temp[i] == '"') {
			if (temp[i+1] == '"') {
				/* Skip past double quote */
				i++;
			} else {
				on_quote = 0;
			}
		} else if (!on_quote && temp[i] == ',') {
			num_cols++;
		}
	}

	/* -- Allocate output array -- */
	ret = malloc(num_rows * sizeof(*ret));
	for (i=0; i<num_rows; i++) {
		ret[i] = malloc(num_cols * sizeof(*ret[i]));
		memset(ret[i], 0, num_cols * sizeof(*ret[i]));
	}

	/* -- Parse it row by row -- */
	for (i=0; i<num_rows; i++) {
		temp = rows[i];
		templen = strlen(temp);
		on_col = 0;
		on_quote = 0;
		ret[i][on_col++]=temp;
		for (j=0; j<templen; j++) {
			if (!on_quote && temp[j] == '"') {
				on_quote = 1;
			} else if (on_quote && temp[j] == '"') {
				if (temp[j+1] == '"') {
					/* Skip past double quote */
					j++;
				} else {
					on_quote = 0;
				}
			} else if (!on_quote && temp[j] == ',') {
				temp[j] = 0;

				/* Sanity check for uneven column counts */
				if (on_col < num_cols) {
					if (ret[i][on_col-1][0] == '"') {
						/* Strip surrounding quotes */
						(ret[i][on_col-1])++;
						ret[i][on_col-1][strlen(ret[i][on_col-1])-1] = 0;
						/* Remove double quotes inside */
						UnescapeQuotes(ret[i][on_col-1]);
					}
					ret[i][on_col++] = temp + (j+1);
				}
			}
		}
		if (ret[i][on_col-1][0] == '"') {
			/* Strip surrounding quotes */
			(ret[i][on_col-1])++;
			ret[i][on_col-1][strlen(ret[i][on_col-1])-1] = 0;
			/* Remove double quotes inside */
			UnescapeQuotes(ret[i][on_col-1]);
		}
	}

	free(rows);

	*num_rows_out = num_rows;
	*num_cols_out = num_cols;
	return(ret);
}
