/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#ifndef __LIBM_MAIN_H__
#define __LIBM_MAIN_H__

#include "monetra.h"

/* Standard includes ... all platforms */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#include <sys/stat.h>

/* Win32 includes */
#ifdef _WIN32
#include <io.h>
#include <winsock2.h>
#include <sys/timeb.h>
#include <windows.h>
#else /* !Win32 */
/* UNIX/POSIX includes */
#include <unistd.h>
#include <sys/socket.h>
#ifndef DISABLE_IP
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <sys/time.h>
#include <dirent.h>

#endif /* Win32 */

typedef struct M_TRANS {
  char *key;
  char *value;
} M_TRANS;

typedef struct M_QUEUE {
  char identifier[31];
  int type;
  int admin;
  int status;

  /* Store outgoing transaction fields -- key/value pairs */
  int transaction_fields;
  M_TRANS *transaction;

  /* Stuff read from return */
  int code;
  int avs;
  int cv;
  char *auth;
  char *text;
  M_int64 tid;
  char *item;
  char *batch;
  
  int resp_fields;
  M_TRANS *parsed_resp;
  /* Comma Delimited response storage */
  char *response;
  int iscommadelimited;
  /* storage */
  char ***separated;
  int columns;
  int rows;
  
  void *next;
  void *prev;
} M_QUEUE;

#include "libmonetra_utils.h"
#include "libmonetra_misc.h"
#include "libmonetra_cdparsing.h"
#include "libmonetra_locking.h"
#include "libmonetra_sockets.h"
#include "libmonetra_file.h"
#include "libmonetra_strings.h"
#include "libmonetra_snprintf.h"
#include "proxy.h"

#ifdef _WIN32
#define ioctl(a,b,c) ioctlsocket(a,b,c)
#ifdef _WIN64
#define fstat(a,b) _fstat64(a,b)
#define stat _stat64
#else
#define fstat(a,b) _fstat(a,b)
#define stat _stat
#endif
#define open(a,b) _open(a,b)
#define read(a,b,c) _read(a,b,c)
#define write(a,b,c) _write(a,b,c)
#define close(a) _close(a)
#define strdup(a) _strdup(a)
#endif

#ifdef M_DROPFILE_ENABLED
#define M_FILE 1
#endif
#define M_SOCKETS 2
#define M_SSL 3



#define M_DISCONNECTED 0
#define M_CONNECTED 1

typedef struct _M_CONN {
  int conn_method;  /* Drop File, Sockets, SSL */
  char location[251]; /* Either Drop-File directory, or hostname */
  unsigned short port;  /* For IP and SSL, a port number */
  int ptr;  /* For IP and SSL, the pointer to an open connection */
  char *inbuf;
  int inbuf_cnt;
  int inbuf_alloc;
  int last_parse_attempt_len;  /* Optimization for large response queries */

  char *outbuf;
  int outbuf_cnt;
  int outbuf_alloc;
  
  int status;
  int do_debug;
  
  long timeout;
  int blocking;
  int verifyconn;
  int verifyssl;
  int validate_identifier;
  int max_conn_time;
  char *error_text;

#ifdef ENABLE_SSL
  void *ssl;
  void *server_cert;
  void *client_ctx;
  char ca_location[255];
  char sslkeyfile[255];
  char sslcertfile[255];
  
#endif
  /* int CA_verify; */
  /* char *CA_cert; */  /* To store an SSL certificate */
  /* char *CA_key; */ /* To store an SSL key */
  long outstanding_auths;
  long queue_length;
  M_QUEUE *queue;

  M_Register_Mutex mutexreg;
  M_Mutex_Lock mutexlock;
  M_Mutex_Unlock mutexunlock;
  M_Unregister_Mutex mutexunreg;
  M_ThreadID threadid;

  void *mutex;
  proxy_config_st proxy;
} _M_CONN;


#define IP_BUFFER_CHUNK_SIZE	128*1024
#define IP_READ_SIZE		1024

#endif
