/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#include "libmonetra_main.h"
#include "monetra.h"

#ifdef ENABLE_SSL
#  include <openssl/rand.h>
#endif

static int M_rand_seeded = 0;

long M_GenerateSeed ()
{
#ifdef _WIN32
	struct _timeb timebuffer;
	_ftime (&timebuffer);
	return ((long)((M_int64)timebuffer.millitm + ((M_int64)1000000 * (M_int64)timebuffer.time)));
#else
	struct timeval t;
	gettimeofday (&t, NULL);
	return (t.tv_usec + (1000000 * t.tv_sec));
#endif
}

int M_DirectoryExists (const char *directory)
{
#ifdef _WIN32
	if ((_access (directory, 0)) != -1)
		return (0);
#else
	DIR *dir;
	dir = opendir (directory);
	if (dir == NULL)
		return (0);
	closedir (dir);
#endif
	return (1);
}

void M_FixPath_inplace(char *string)
{
	int len, i;
	int offset = 0;
	if (string == NULL) return;
	len = strlen(string);
	for (i=0; i<len; i++) {
		if (string[i] == '/' || string[i] == '\\') {
			if (i != 0 && (string[i-1] == '/' && string[i-1] == '\\')) {
				offset++;
			} else {
#ifdef _WIN32
				string[i-offset] = '\\';
#else
				string[i-offset] = '/';
#endif
			}
		} else {
			string[i-offset] = string[i];
		}
	}
	string[i-offset]=0;
}

char *M_midstr (const char *string, int start, int length)
{
	const char *temp = NULL;
	char *ret = NULL;
	temp = string + start;
	ret = (char *) malloc (length + 1);

	if (ret == NULL)
		return (NULL);
	memset(ret, 0, length+1);
	strncpy(ret, temp, length);

	return (ret);
}

/* Generate a random number which can be used in a filename if necessary */
static int M_RandSeed(void)
{
	long seed;
	seed = M_GenerateSeed ();
	srand ((unsigned int) seed);
	M_rand_seeded = 1;
	return (1);
}

static int M_RandChar(void)
{
	int rand_num;
	int done = 0;

	while (!done) {
		rand_num = 48 + rand () % 74;
		if ((rand_num >= 58 && rand_num <= 63)
		    || (rand_num >= 92 && rand_num <= 96))
			done = 0;
		else
			done = 1;
	}
	return (rand_num);
}

char *M_GenerateIdentifier ()
{
	char *identifier;
	int i;
	unsigned char rand_bytes[12];
	int use_rand=0;

#ifdef ENABLE_SSL
	/* Use real random bytes if possible */
	memset(rand_bytes, 0, sizeof(rand_bytes));
	if (RAND_bytes(rand_bytes, sizeof(rand_bytes)))
		use_rand = 1;
#endif

	identifier = (char *) malloc(25);
	memset(identifier, 0, 25);
	if (use_rand) {
		for (i=0; i<12; i++) {
			identifier[i*2]    = ("0123456789ABCDEF")[0x0f & (rand_bytes[i] >> 4)];
			identifier[(i*2)+1]= ("0123456789ABCDEF")[0x0f & rand_bytes[i]];
		}
	} else {
		if (!M_rand_seeded) M_RandSeed();
		for (i = 0; i < 20; i++) {
			identifier[i] = M_RandChar ();
		}
	}
	return (identifier);
}


/* Replacement for sleep() and usleep() */
int M_SPEC M_uwait (unsigned long length)
{
#ifdef WIN32
	Sleep(length/1000);
#else
	fd_set readfs;
	struct timeval timeout;

	timeout.tv_sec = length / (unsigned long) 1000000;
	timeout.tv_usec = (length % (unsigned long) 1000000);
	FD_ZERO (&readfs);
	select (0, &readfs, NULL, NULL, &timeout);
#endif
	return (1);
}

char *MC_SAFE_strncpy (char *dest, const char *src, long len)
{
	if (src == NULL || !strlen (src)) {
		dest[0] = 0;
	} else {
		strncpy (dest, src, len);
	}
	return (dest);
}

char *MC_SAFE_strdup (const char *src)
{
	char *ret = NULL;
	if (src == NULL || !strlen (src)) {
		ret = malloc (1);
		ret[0] = 0;
	} else {
		ret = strdup (src);
	}
	return (ret);
}



void M_Set_Conn_Error (M_CONN * myconn, const char *error)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	if (error != NULL && strlen (error)) {
		if (conn->error_text != NULL)
			free (conn->error_text);
		conn->error_text = NULL;
		conn->error_text = strdup (error);
	}
}
