/* Monetra 5.x C API
For any questions please contact  support@mainstreetsoftworks.com

Copyright (c) 2005, Main Street Softworks, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the Main Street Softworks, Inc. nor the names of its
      contributors may be used to endorse or promote products derived from 
      this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#include "monetra.h"
#include "libmonetra_main.h"
#include "proxy.h"

#ifdef ENABLE_SSL
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#endif

#ifndef _WIN32
#include <netdb.h>
#endif
#ifndef DISABLE_IP

int M_InitSockets ()
{
#ifdef _WIN32
	WSADATA wsaData;

	if (WSAStartup (0x202, &wsaData) == SOCKET_ERROR) {
		fprintf (stderr, "WSAStartup failed with error %d\n",
			 WSAGetLastError ());
		WSACleanup ();
		return 0;
	}
#endif /* _WIN32 */
	return (1);
}

void M_closesocket(int fd, int call_shutdown)
{
	if (call_shutdown) {
		M_SetNonBlock(fd, 0); /* Make sure we're in blocking for shutdown() */
#ifdef _WIN32
		shutdown(fd, SD_BOTH);
#else
#  if !defined(SHUT_RDWR) && defined(SHUT_RD)
#    define SHUT_RDWR SHUT_RD
#  elif !defined(SHUT_RDWR)
#    define SHUT_RDWR 2
#  endif
		shutdown(fd, SHUT_RDWR);
#endif
		M_uwait(10000); /* Wait arbitrary time for shutdown */
	}

#ifdef _WIN32
	closesocket(fd);
#else
	close(fd);
#endif
}

int M_SetNonBlock(int fd, int tf)
{
#ifdef O_NONBLOCK
	int flags;

	if ((flags=fcntl(fd, F_GETFL, 0)) == -1)
		return(0);
	if (tf)
		flags |= O_NONBLOCK;
	else
		flags &= ~(O_NONBLOCK);

	if (fcntl(fd, F_SETFL, flags) == -1)
		return(0);
#else
#ifndef __WATCOMC__
	if (ioctl(fd, FIONBIO, &tf) == -1)
#else
	if (ioctl(fd, FIONBIO, (char *)&tf) == -1)
#endif
		return(0);
#endif
	return(1);
}

long M_read_socket(const char *dbg_head, int dbg_level, char *buf, int buflen, int fd)
{
	(void)dbg_head;
	(void)dbg_level;
#ifdef _WIN32
	return (recv (fd, buf, buflen, 0));
#else
	return (read (fd, buf, buflen));
#endif
}

long M_write_socket(const char *dbg_head, int dbg_level, char *buf, int buflen, int fd)
{
	(void)dbg_head;
	(void)dbg_level;
#ifdef _WIN32
	return (send (fd, buf, buflen, 0));
#else
	return (write (fd, buf, buflen));
#endif
}

#ifndef INADDR_NONE
#define INADDR_NONE 0xFFFFFFFF
#endif

int M_ip_connect(M_CONN * myconn)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	struct sockaddr_in peer;
	int soct;
	int len;
	char *addr=NULL;
	int ip_connect_timeout = 5;
	char ipaddr[50];
	char error[255];
	char temperr[255];
#ifdef _WIN32
	const char *w32_addr=NULL;
#endif
	proxy_config_st *proxy = NULL;
	time_t starttime, t;

	proxy = &(conn->proxy);
	ip_connect_timeout = conn->max_conn_time;

	time(&starttime);
	if (PROXY_NS(proxy_tunnel_check)(proxy, conn->location, conn->port)) {
		soct = PROXY_NS(proxy_tunnel_connect)(proxy, conn->location, conn->port, error, sizeof(error));
		if (soct == -1) {
			M_snprintf(temperr, sizeof(temperr), "Proxy Connect Failed: %s", error);
			M_Set_Conn_Error (myconn, temperr);
			return(-1);
		}
	} else {
		memset(&peer, 0, sizeof(peer));
		peer.sin_family = AF_INET;
		peer.sin_port = htons(conn->port);

		memset(temperr, 0, sizeof(temperr));
		len=PROXY_NS(ghbn)(conn->location, &addr, error, sizeof(error));
		if (len != -1) {
			memcpy((char *)&peer.sin_addr, addr, len);
			free(addr);
		} else if (strcasecmp (conn->location, "localhost") == 0) {
			free(addr);
			peer.sin_addr.s_addr = inet_addr("127.0.0.1");
		} else {
			free(addr);
			M_snprintf(temperr, sizeof(temperr), "DNS Lookup Failed: %s", error);
			M_Set_Conn_Error (myconn, temperr);
			return(-1);
		}

		memset(ipaddr, 0, sizeof(ipaddr));
#ifdef _WIN32
		w32_addr = inet_ntoa(peer.sin_addr);
		if (w32_addr != NULL)
			strncpy(ipaddr, w32_addr, sizeof(ipaddr));
#else
		inet_ntop(AF_INET, (const void *)&peer.sin_addr, ipaddr, sizeof(ipaddr));
#endif

		time(&t);
		starttime = t;
		soct = socket(AF_INET, SOCK_STREAM, 0);
		if (soct < 0) {
			M_snprintf(error, sizeof(error), "socket() failed, could not get fd: %s", strerror(errno));
			M_Set_Conn_Error (myconn, error);
			return(-1);
		}

		if (!PROXY_NS(connect_nonblock)(soct, (struct sockaddr *)&peer, sizeof(peer), ip_connect_timeout, error, sizeof(error))) {
			M_closesocket(soct, 0);
			M_snprintf(temperr, sizeof(temperr), "IP Connection failed: %s", error);
			M_Set_Conn_Error (myconn, temperr);
			return(-1);
		}
	}
	time(&t);

	return(soct);
}

int M_VerifyPing (M_CONN * myconn)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	M_uintptr identifier = -1;
	time_t t, lastt;
	int status = 0;
	int blocking = 0;

	blocking = conn->blocking;

	M_SetBlocking (myconn, 0);

	identifier = M_Ping (myconn);
	time (&lastt);
	t = lastt;
	status = 1;
	while (M_CheckStatus (myconn, identifier) != M_DONE) {
		if (!M_Monitor (myconn)) {
			status = 0;
			break;
		}
		time (&t);
		if ((t - lastt) > 3) {	/* Response Timeout */
			status = 0;
			break;
		}
		M_uwait (10000);
	}

	M_SetBlocking (myconn, blocking);

	M_DeleteResponse (myconn, identifier);
	if (!status) {

		return (0);
	}
	return (1);
}

/* Let's see if there's anything in the buffer to read! */
int M_CheckRead_select(int ptr, long delay)
{
	fd_set readfs;
	struct timeval timeout;
	int res;

	FD_ZERO (&readfs);
	FD_SET (ptr, &readfs);
	timeout.tv_sec = 0;
	timeout.tv_usec = delay;
	res = select (ptr + 1, &readfs, NULL, NULL, &timeout);
	if (res > 0) {
		if (FD_ISSET (ptr, &readfs)) {
			return (1);
		}
	}
	return (0);
}


/* Let's make sure that the buffer isn't too full to write to
    I'd assume this will only happen in dial-up scenerios */
int M_CheckWrite_select(int ptr, long delay)
{
	fd_set sendfs;
	struct timeval timeout;
	int res;

	FD_ZERO (&sendfs);
	FD_SET (ptr, &sendfs);
	timeout.tv_sec = 0;
	timeout.tv_usec = delay;

	res = select (ptr + 1, NULL, &sendfs, NULL, &timeout);
	if (res > 0) {
		if (FD_ISSET (ptr, &sendfs)) {
			return (1);
		}
	}
	return (0);
}


#ifndef _WIN32
static FILE *M_openlog(void)
{
	char debug_file[255];
	FILE *fp=NULL;
	M_snprintf(debug_file, sizeof(debug_file), "/tmp/libmonetra-%d.log", getpid());
	fp = fopen(debug_file, "ab");
	return(fp);
}

static void M_output_buffer(FILE *fp, const char *prefix, const char *buf, int len)
{
	int i,c;
	if (fp == NULL) return;
	for (i=0; i<len; i++) {
		c=buf[i];
		if (c < 32 || c > 126) c=32;
		fprintf(fp, "%s: %05d : %c -- HEX : 0x%0x DEC: %02d\n", prefix, len, c, buf[i], buf[i]);
	}
}

#endif

/* Helper macro to efficiently append data to the end of a buffer and allocate
 * in blocks */
#define M_append_ipssldata(buffer, buffer_len, buffer_alloc, append, append_len) \
	do { \
		if (buffer_len + append_len > buffer_alloc) { \
			buffer_alloc = (((buffer_len + append_len) / (IP_BUFFER_CHUNK_SIZE)) + 1) * (IP_BUFFER_CHUNK_SIZE); \
			buffer = realloc(buffer, buffer_alloc); \
		} \
		memcpy(buffer + buffer_len, append, append_len); \
		buffer_len += append_len; \
	} while(0)

int M_Monitor_IP(M_CONN *myconn, unsigned long delay)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	char temp_buffer[IP_READ_SIZE];
#ifndef _WIN32
	FILE *fp = NULL;
#endif
	int bytes_read = 0;
	int bytes_written = 0;
	int status = 1;
	int outlen;
	int loopcnt;

	size_t len;
	len = 0;
#ifndef _WIN32
	if (conn->do_debug) {
		fp = M_openlog(); 
	}
#endif

#ifndef _WIN32
	if (conn->do_debug && fp != NULL) {
		fprintf (fp, "Looking to write\n");
	}
#endif
	M_lock (myconn);
	if (status && conn->outbuf_cnt != 0) {
		if (M_CheckWrite_select(conn->ptr, 0)) {
			outlen = conn->outbuf_cnt;
			if (conn->conn_method == M_SOCKETS) {
				bytes_written = M_write_socket(NULL, 0, conn->outbuf, outlen, conn->ptr);
			} else if (conn->conn_method == M_SSL) {
#ifdef ENABLE_SSL
				bytes_written = SSL_write ((SSL *) conn->ssl, conn->outbuf, outlen);
				if (bytes_written <= 0) {
					/* connection closed */
					status = 0;
				}
#endif
			}
#ifndef _WIN32
			if (conn->do_debug && fp != NULL) {
				fprintf (fp, "Wrote %d bytes\n", outlen);
				M_output_buffer(fp, "WROTE", conn->outbuf, conn->outbuf_cnt);
			}
#endif
			if (bytes_written < conn->outbuf_cnt) {
				memmove (conn->outbuf,
					 conn->outbuf + bytes_written,
					 conn->outbuf_cnt - bytes_written);
				conn->outbuf_cnt -= bytes_written;
			} else {
				free (conn->outbuf);
				conn->outbuf = NULL;
				conn->outbuf_cnt = 0;
				conn->outbuf_alloc = 0;
			}
		}
	}

	M_unlock (myconn);

#ifndef _WIN32
	if (conn->do_debug && fp != NULL) {
		fprintf (fp, "Looking to read\n");
	}
#endif
	M_lock (myconn);
	loopcnt = 0;
	while (status && (
#ifdef ENABLE_SSL
		(conn->conn_method == M_SSL && SSL_pending((SSL *) conn->ssl))
#endif
		|| M_CheckRead_select(conn->ptr, delay))) {

		if (conn->conn_method == M_SOCKETS) {
			bytes_read = M_read_socket(NULL, 0, temp_buffer, IP_READ_SIZE, conn->ptr);
		} else if (conn->conn_method == M_SSL) {
#ifdef ENABLE_SSL
			bytes_read = SSL_read ((SSL *) conn->ssl, temp_buffer, IP_READ_SIZE);
#endif
		}

#ifndef _WIN32
		if (conn->do_debug && fp != NULL) {
			fprintf (fp, "Read %d bytes\n", bytes_read);
			M_output_buffer(fp, "READ", temp_buffer, bytes_read);
		}
#endif
		if (bytes_read <= 0) {
			/* connection closed */
			status = 0;
			break;
		}

		M_append_ipssldata(conn->inbuf, conn->inbuf_cnt, conn->inbuf_alloc, temp_buffer, bytes_read);
#ifndef _WIN32
		if (conn->do_debug && fp != NULL) {
			fprintf(fp, "Full inbuffer size: %d\n", conn->inbuf_cnt);
			M_output_buffer(fp, "FULL INBUF", conn->inbuf, conn->inbuf_cnt);
		}
#endif

		loopcnt++;

		/* Don't loop if we didn't fill the buffer */
		if (bytes_read < IP_READ_SIZE || loopcnt > 256) {
			break;
		}
	}
	M_unlock (myconn);

#ifndef _WIN32
	if (conn->do_debug) {
		fclose (fp);
	}
#endif
	return (status);
}

int M_SendTransaction_IP (M_CONN * myconn, char *identifier,
			  char *transaction)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	int ident_len = strlen (identifier);
	int msg_len = strlen (transaction);

	M_lock (myconn);

	M_append_ipssldata(conn->outbuf, conn->outbuf_cnt, conn->outbuf_alloc, "\x02", 1);
	M_append_ipssldata(conn->outbuf, conn->outbuf_cnt, conn->outbuf_alloc, identifier, ident_len);
	M_append_ipssldata(conn->outbuf, conn->outbuf_cnt, conn->outbuf_alloc, "\x1c", 1);
	M_append_ipssldata(conn->outbuf, conn->outbuf_cnt, conn->outbuf_alloc, transaction, msg_len);
	M_append_ipssldata(conn->outbuf, conn->outbuf_cnt, conn->outbuf_alloc, "\x03", 1);

	M_unlock (myconn);
	return (1);
}

#endif /* DISABLE_IP */
