/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#include "libmonetra_main.h"
#ifndef _WIN32
#include <strings.h>
#endif
static M_QUEUE *M_FindQueue (M_CONN * myconn, char *identifier)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	M_QUEUE *ptr = NULL;

	ptr = conn->queue;

	while (ptr != NULL) {
		if (ptr->status == M_PENDING
		    && strcmp (ptr->identifier, identifier) == 0)
			return (ptr);
		ptr = (M_QUEUE *) ptr->next;
		if (ptr == conn->queue)
			break;
	}
	return (NULL);
}

static void M_StoreResponseValues (M_QUEUE * queue, const char *option, const char *value)
{
	if (strcmp (option, "code") == 0) {
		if (strcmp (value, "AUTH") == 0)
			queue->code = M_AUTH;
		else if (strcmp (value, "DENY") == 0)
			queue->code = M_DENY;
		else if (strcmp (value, "DUPL") == 0)
			queue->code = M_DUPL;
		else if (strcmp (value, "CALL") == 0)
			queue->code = M_CALL;
		else if (strcmp (value, "PKUP") == 0)
			queue->code = M_PKUP;
		else if (strcmp (value, "RETRY") == 0)
			queue->code = M_RETRY;
		else if (strcmp (value, "SETUP") == 0)
			queue->code = M_SETUP;
		else if (strcmp (value, "SUCCESS") == 0)
			queue->code = M_SUCCESS;
		else if (strcmp (value, "FAIL") == 0)
			queue->code = M_FAIL;
		else if (strcmp (value, "TIMEOUT") == 0)
			queue->code = M_TIMEOUT;
		else
			queue->code = M_DENY;
	} else if (strcmp (option, "avs") == 0) {
		if (strcmp (value, "GOOD") == 0)
			queue->avs = M_GOOD;
		else if (strcmp (value, "BAD") == 0)
			queue->avs = M_BAD;
		else if (strcmp (value, "STREET") == 0)
			queue->avs = M_STREET;
		else if (strcmp (value, "ZIP") == 0)
			queue->avs = M_ZIP;
		else
			queue->avs = M_UNKNOWN;
	} else if (strcmp (option, "cv") == 0) {
		if (strcmp (value, "GOOD") == 0)
			queue->cv = M_GOOD;
		else if (strcmp (value, "BAD") == 0)
			queue->cv = M_BAD;
		else
			queue->cv = M_UNKNOWN;
	} else if (strcmp (option, "auth") == 0) {
		queue->auth = strdup (value);
	} else if (strcmp (option, "verbiage") == 0) {
		queue->text = strdup (value);
	} else if (strcmp (option, "item") == 0) {
		queue->item = strdup (value);
	} else if (strcmp (option, "batch") == 0) {
		queue->batch = strdup (value);
	} else if (strcmp (option, "ttid") == 0 || strcmp (option, "tid") == 0
		   || strcmp (option, "sid") == 0
		   || strcmp (option, "did") == 0) {
		queue->tid = M_atoll (value);
	}
}

static void M_add_resp_keyval(M_QUEUE *ptr, const char *key, const char *val)
{
	if (ptr->resp_fields % 10 == 0)
		ptr->parsed_resp = realloc(ptr->parsed_resp, (ptr->resp_fields + 10)*sizeof(M_TRANS));
	ptr->parsed_resp[ptr->resp_fields].key = MC_SAFE_strdup(key);
	ptr->parsed_resp[ptr->resp_fields].value = MC_SAFE_strdup(val);
	ptr->resp_fields++;
}

/* Returns NULL if no quoting needed, otherwise allocated
 * NULL-terminated buffer of quoted data */
char *M_encode_quoted(const char *str, int len)
{
	int alloc_len=0;
	int cnt=0, i;
	char *ret=NULL;
	int needs_quoting = 0; /* 0 = no, 1 = yes, 2 = yes and has embedded quotes */

	for (i=0; i<len; i++) {
		switch(str[i]) {
			case '\r':
			case '\n':
			case ',':
				needs_quoting = 1;
			break;
			case '"':
				needs_quoting = 2;
			break;
		}
		if (needs_quoting == 2)
			break;
	}

	if (!needs_quoting)
		return(NULL);

	if (needs_quoting == 2) {
		alloc_len = (len*2) + 2;
	} else if (needs_quoting == 1) {
		alloc_len = len + 2;
	}

	ret=(char *)malloc(alloc_len+1);

	if (needs_quoting == 1) {
		ret[0] = '"';
		memcpy(ret+1, str, len);
		ret[len+1] = '"';
		ret[len+2] = 0;
	} else {
		ret[cnt++] = '"';
		for (i=0; i<len; i++) {
			if (str[i] == '"')
				ret[cnt++]='"';
			ret[cnt++]=str[i];
		}
		ret[cnt++] = '"';
		ret[cnt] = 0;
	}
	return(ret);
}

void UnescapeQuotes(char *quoted_data)
{
	int i, cnt;
	for (i=0, cnt=0 ; quoted_data != NULL && quoted_data[i] != 0; i++) {
		if (quoted_data[i] == '"' && quoted_data[i+1] == '"') {
			/* Skip one of the quotes */
			i++;
		}
		quoted_data[cnt++]=quoted_data[i];
	}
	if (quoted_data) 
		quoted_data[cnt] = 0;
}

void M_ProcessResponse (M_CONN * myconn, char *identifier, char *message)
{
	int len, i;
	char *key = NULL, *val = NULL;
	int on_quote = 0; /* 0 = not on quote, 1 = on quote, 2 = quote done (wait till end of line to store) */
	int all_whitespace = 0;
	int multi_quote = 0;
	int vallen, j;

	M_QUEUE *ptr = NULL;

	/* Must already be holding a lock on myconn!! */
	ptr = M_FindQueue (myconn, identifier);
	if (ptr == NULL) {
		return;
	}
	ptr->status = M_DONE;

	len = strlen(message);
	/* See if it's comma delimited */
	ptr->iscommadelimited = 1;
	for (i=0; i<len; i++) {
		if (message[i] == '=') {
			ptr->iscommadelimited=0;
			break;
		}
		if (i >= 25 || message[i] == '\r' || message[i] == '\n')
			break;
	}

	if (ptr->iscommadelimited) {
		len = strlen (message);
		ptr->response = (char *) malloc (len + 1);
		memcpy (ptr->response, message, len);
		ptr->response[len] = 0;

		ptr->code = M_SUCCESS;
	} else {

		/* Parse out key/value pairs from each line */
		key = message;
		for (i=0; i<len; i++) {
			if (on_quote != 1 && (message[i] == '\r' || message[i] == '\n')) {
				message[i] = 0;
				if (key != NULL && val != NULL) {
					if (on_quote) {
						/* Strip leading and trailing quotes */
						val++;
						val[strlen(val)-1]=0;
						if (multi_quote) {
							UnescapeQuotes(val);
						}
					}
					M_add_resp_keyval(ptr, key, val);
					M_StoreResponseValues(ptr, key, val);
					key = NULL;
					val = NULL;
					on_quote = 0;
					multi_quote = 0;
				}
			} else if (on_quote == 0 && val != NULL && message[i] == '"') {
				/* If on value, and we hit a quote, lets see if it comes before
				 * any non-whitespace values */
				all_whitespace = 1;
				/* Current length of value up to this point */
				vallen = val - (message+i) - 1;
				for (j=0; j<vallen; j++) {
					if (val[j] != ' ' && val[j] != '\t') {
						all_whitespace = 0;
						break;
					}
				}
				/* Ok, properly quoted, treat as such */
				if (all_whitespace) {
					on_quote = 1;
					val = message + i; /* New val start is at quote */
				}
			} else if (on_quote == 1 && message[i] == '"') {
				if (message[i+1] == '"') {
					/* Quoted quotes in this value */
					multi_quote = 1;
					i++; /* Skip over next quote */
				} else {
					/* Found ending quote */
					on_quote=2;

					/* Null terminate after ending quote, unless next
					 * character is a newline */
					if (message[i+1] != '\r' && message[i+1] != '\n')
						message[i+1] = 0;
				}
			} else if (message[i] == '=' && val == NULL) {
				message[i] = 0;
				val = message+(i+1);
			} else if (key == NULL) {
				key = message+i;
			}
		}
		if (key != NULL && val != NULL) {
			if (on_quote) {
				/* Strip leading and trailing quotes */
				val++;
				val[strlen(val)-1]=0;
				if (multi_quote) {
					UnescapeQuotes(val);
				}
			}
			M_add_resp_keyval(ptr, key, val);
			M_StoreResponseValues(ptr, key, val);
		}
	}

}

static int M_IP_GetTransParams (const char *data, int len, int last_parse_attempt_len,
				M_uintptr *start, M_uintptr *fs, M_uintptr *finish)
{
	const char *end_ptr, *fs_ptr = NULL;

	(*start) = 0;
	(*fs) = 0;
	(*finish) = 0;

	if (data == NULL || len < 3)
		return 0;

	if (last_parse_attempt_len > len) {
		printf("ERROR -- last_parse_attempt_len > len, invalid\r\n");
		last_parse_attempt_len = 0;
	}

	/* Find end first, as fast as possible, as it may not be
	 * there yet */
	end_ptr = memchr(data + last_parse_attempt_len, 0x03, len - last_parse_attempt_len);
	if (end_ptr == NULL)
		return 0;

	(*finish) = (M_uintptr)(end_ptr - data);

	if (data[0] != 0x02) {
		(*start) = 1;
		return -1;
	}

	(*start) = 0;

	fs_ptr = memchr(data+1, 0x1c, (*finish)-2);
	if (fs_ptr == NULL) {
		/* Umm, should never get here, invalid message */
		return 0;
	}

	(*fs) = (M_uintptr)(fs_ptr - data);

	return (1);
}

static void M_debug_outputbuffer(const char *buffer, int len)
{
	int i, c;
	for (i=0; i<len; i++) {
		c=buffer[i];
		if (c >=32 && c<=126) {
		} else {
			c=32;
		}
		printf("BUFFER: %05d: %c -- HEX: 0x%02X DEC: %02d\r\n", i, c, buffer[i], buffer[i]);
	}
}

/* return 1 on success, return 0 on critical failure */
int M_ProcessBuffer (M_CONN * myconn)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	int parse_offset = 0;
	M_uintptr parse_start = 0, parse_fs = 0, parse_finish = 0;
	char *identifier = NULL, *message = NULL;

	M_lock (myconn);

	if (conn->inbuf != NULL) {
		parse_offset = 0;
		while (parse_offset < conn->inbuf_cnt
		       && M_IP_GetTransParams (conn->inbuf + parse_offset,
					       conn->inbuf_cnt - parse_offset,
					       conn->last_parse_attempt_len,
					       &parse_start, &parse_fs,
					       &parse_finish)) {

			/* Reset this so that it will rescan the entire
			 * buffer, rather than the partial buffer, as this
			 * flag is an optimization for _large_ reports */
			conn->last_parse_attempt_len = 0;

			if (parse_start != 0) {
				M_unlock (myconn);
				printf("error, first character of buffer is NOT STX. Either bad response from Monetra, or memory corruption from integrated app!\r\n");
				printf("Parse offset: %d of %d bytes\r\n", parse_offset, conn->inbuf_cnt);
				M_debug_outputbuffer(conn->inbuf, conn->inbuf_cnt);
				return(0);
			//	printf ("error error, should never get here!\r\n");
			//	return;
			}
			conn->inbuf[parse_offset + parse_fs] = 0;
			conn->inbuf[parse_offset + parse_finish] = 0;
			identifier = conn->inbuf + (parse_offset + parse_start + 1);
			message = conn->inbuf + (parse_offset + parse_fs + 1);
			M_ProcessResponse(myconn, identifier, message);
			parse_offset += (parse_finish + 1);
		}
		if (parse_offset < conn->inbuf_cnt) {
			memmove (conn->inbuf, conn->inbuf + parse_offset,
				 conn->inbuf_cnt - parse_offset);
			conn->inbuf_cnt -= parse_offset;
		} else {
			free (conn->inbuf);
			conn->inbuf = NULL;
			conn->inbuf_cnt = 0;
			conn->inbuf_alloc = 0;
		}
		/* How many bytes had already been scanned for a complete
		 * transaction on the last attempt? */
		conn->last_parse_attempt_len = conn->inbuf_cnt;
	}

	M_unlock (myconn);
	return(1);
}

M_QUEUE *M_NewQueueSlot (M_CONN * myconn)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	M_QUEUE *ptr = NULL, *base = NULL, *last = NULL;

	ptr = (M_QUEUE *) malloc (sizeof (M_QUEUE));
	ptr->next = NULL;
	ptr->prev = NULL;
	ptr->status = M_UNUSED;
	ptr->identifier[0] = 0;
	ptr->type = 0;
	ptr->admin = 0;
	ptr->status = M_NEW;
	ptr->transaction_fields = 0;
	ptr->transaction = NULL;
	ptr->code = -1;
	ptr->avs = -1;
	ptr->cv = -1;
	ptr->auth = NULL;
	ptr->text = NULL;
	ptr->tid = -1;

	ptr->item = NULL;
	ptr->batch = NULL;

	ptr->resp_fields = 0;
	ptr->parsed_resp = NULL;
	ptr->response = NULL;
	ptr->iscommadelimited = 0;

	ptr->separated = NULL;
	ptr->columns = 0;
	ptr->rows = 0;

	base = conn->queue;
	/* Circular queue! */
	if (base == NULL) {
		conn->queue = ptr;
		ptr->next = ptr;
		ptr->prev = ptr;
	} else {
		last = base->prev;
		last->next = ptr;
		ptr->prev = last;
		ptr->next = base;
		base->prev = ptr;
	}

	conn->queue_length++;

	return (ptr);
}

#define TRANS_BLOCK_SIZE 4096
#define trans_append_block(ret, real_size, alloc_size, data, len) \
		do { \
			if (real_size + len + 1 > alloc_size) { \
				alloc_size = (((real_size + len + 1) / TRANS_BLOCK_SIZE) + 1) * TRANS_BLOCK_SIZE; \
				ret = (char *)realloc(ret, alloc_size); \
			} \
			memcpy(ret + real_size, data, len); \
			real_size += len; \
			ret[real_size] = 0; \
		} while(0)
char *M_StructureTransaction (M_CONN * myconn, M_QUEUE * queue)
{
	char *transaction = NULL;
	char *temp=NULL;
	int alloc_len=0;
	int len=0, templen=0;
	int i;

	(void)myconn;

	if (queue->type == MC_TRAN_PING) {
		transaction = malloc(7);
		M_snprintf(transaction, 7, "%s", "PING\r\n");
		return (transaction);
	}
	for (i = 0; i < queue->transaction_fields; i++) {
		templen = strlen(queue->transaction[i].key);
		trans_append_block(transaction, len, alloc_len, queue->transaction[i].key, templen);
		trans_append_block(transaction, len, alloc_len, "=", 1);
		if (queue->transaction[i].value) {
			templen = strlen(queue->transaction[i].value);
			temp = M_encode_quoted(queue->transaction[i].value, templen);
			if (temp == NULL) {
				trans_append_block(transaction, len, alloc_len, queue->transaction[i].value, templen);

			} else {
				templen = strlen(temp);
				trans_append_block(transaction, len, alloc_len, temp, templen);
				free(temp);
			}
		}
		trans_append_block(transaction, len, alloc_len, "\r\n", 2);
	}
	return (transaction);
}

int M_SendTransaction (M_CONN * myconn, M_QUEUE * queue)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	char *transaction = NULL;
	char *id = NULL;
	int ret = 0;

	id = M_GenerateIdentifier ();
	transaction = M_StructureTransaction (myconn, queue);

#ifdef M_DROPFILE_ENABLED
	if (conn->conn_method == M_FILE)
		ret = M_SendTransaction_File (myconn, id, transaction);
#endif
#ifndef DISABLE_IP
	if (conn->conn_method == M_SOCKETS || conn->conn_method == M_SSL)
		ret = M_SendTransaction_IP (myconn, id, transaction);
#endif

	free (transaction);

	// pos=M_AddOutstandingAuth(myconn, identifier);
	M_lock (myconn);
	queue->status = M_PENDING;
	strcpy (queue->identifier, id);
	free (id);
	M_unlock (myconn);

	return (ret);
}

int M_verify_trans_in_queue(M_CONN * myconn, M_uintptr identifier)
{
	_M_CONN *conn = (_M_CONN *) (*myconn);
	void *trans=(void *)identifier;
	int ret=0;
	M_QUEUE *queue=NULL, *ptr=NULL;

	if (myconn == NULL)
		return 0;

	M_lock(myconn);
	if (conn->validate_identifier) {
		queue=conn->queue;
		ptr=queue;
		while (ptr != NULL) {
			if (ptr == trans) {
				ret=1;
				break;
			}
			ptr=ptr->next;
			/* Don't remember if it's circular .. but just in case */
			if (ptr == queue) break;
		}
	} else {
		ret=1;
	}
	M_unlock(myconn);
	return(ret);
}
